
// Operations on collapsed ranges.


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CollapsedRangeRep': collapsed template range.                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class CollapsedRangeRep : public RangeRep {
public :
  CollapsedRangeRep(const int _extent) :
      RangeRep(_extent, 0, 1, Dimension(), DIST_COLLAPSED) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  RangeRep* ker() const ;

  RangeRep* shell() const ;

  void location(Location* loc, const int glb) const ;

  void block(const int ker_glb, Block* blk) const ;

  RangeRep* subker() const ;

  int volume() const ;

  void show() ;

protected:
  int parSize ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CollapsedSubRangeRep': collapsed subrange.                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class CollapsedSubRangeRep : public SubRangeRep {
public :
  CollapsedSubRangeRep(const int _extent,
		       const int _base, const int _stride,
		       RangeRep* _parent, const int _parSize) :
          SubRangeRep(_extent, _base, _stride, _parent, 0, 1, Dimension(),
                      DIST_COLLAPSED),
          parSize(_parSize) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  RangeRep* ker() const ;

  RangeRep* shell() const ;

  void location(Location* loc, const int glb) const ;

  void block(const int ker_glb, Block* blk) const ;

  RangeRep* subker() const ;

  int volume() const ;

  int glb(const int sub, const int ker_glb) const ;

  void show() ;

protected :

  int parSize ;

MEMORY_POOL_DECS
} ;

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ExtendedCollapsedRangeRep',                                //
//         `ExtendedCollapsedSubRangeRep':                             //
//     Collapsed range of array with ghost areas.                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ExtendedCollapsedRangeRep : public CollapsedRangeRep {
public :
  ExtendedCollapsedRangeRep(const int _extent,
                            const int _wlo, const int _whi) :
          CollapsedRangeRep(_extent), wlo(_wlo), whi(_whi) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  int disp(const int sub) const ;

  int volume() const ;

protected :
  int wlo, whi ;
 
MEMORY_POOL_DECS
} ;


class ExtendedCollapsedSubRangeRep : public CollapsedSubRangeRep {
public :
  ExtendedCollapsedSubRangeRep(const int _extent,
                           const int _base, const int _stride,
                           RangeRep* _parent, const int _parSize,
                           const int _wlo, const int _whi) :
    CollapsedSubRangeRep(_extent, _base, _stride, _parent, _parSize),
    wlo(_wlo), whi(_whi) {}

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  int disp(const int sub) const ;

  int volume() const ;

protected :
  int wlo, whi ;

MEMORY_POOL_DECS
} ;


//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
////                                                                     //
//// class `PackedCollapsedSubRangeRep':                                 //
////      Collapsed range of array with ``packed'' layout.               //
////                                                                     //
//// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//
//class PackedCollapsedSubRangeRep : public CollapsedSubRangeRep {
//public :
//  PackedCollapsedSubRangeRep(const int _extent,
//                             const int _base, const int _stride,
//                             CollapsedRangeRep* _parent,
//                             RangeRep* _kernel, const int _parSize,
//                             const int _packing) :
//        CollapsedSubRangeRep(_extent, _base, _stride, _parent,
//                             _kernel, _parSize),
//        packing(_packing) {}
//
//  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;
//
//  int disp(const int sub) const ;
//  int step(const int sub_stp) const ;
//
//  int width() const ;
//
//protected :
//  int packing ;
//
//MEMORY_POOL_DECS
//} ;



// -------------------------------------------------------------
// Things to do
//

