
// Operations on cyclic distributed ranges.


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CyclicRangeRep': Cyclic-distributed template range.          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class CyclicRangeRep : public RangeRep {
public :
  CyclicRangeRep(const int _extent, RangeRep* _kernel, const int _numLayers) :
          RangeRep(_extent, _kernel, _kernel->lev() + 1, _kernel->dim(),
                   DIST_CYCLIC),
          numLayers(_numLayers), numProcs(_kernel->size()) {
    kernel->incr() ;
  }

  ~CyclicRangeRep() {
    kernel->decr() ;
  }

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  RangeRep* shell() const ;

  void location(Location* loc, const int glb) const ;

  void block(const int ker_glb, Block* blk) const ;

  int volume() const ;

  int glb(const int sub, const int ker_glb) const ;

  void show() ;

protected :
  int numLayers, numProcs ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CyclicSubRangeRep': cyclic-distributed subrange.             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class CyclicSubRangeRep : public SubRangeRep {
public :
  CyclicSubRangeRep(const int _extent,
		    const int _base, const int _stride,
		    RangeRep* _parent,
		    RangeRep* _kernel, const int _numLayers) :
          SubRangeRep(_extent, _base, _stride, _parent,
                      _kernel, _kernel->lev() + 1, _kernel->dim(),
                      DIST_CYCLIC),
          numLayers(_numLayers), numProcs(_kernel->size()) {
    euler() ;
  }

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  RangeRep* shell() const ;

  void location(Location* loc, const int glb) const ;

  void block(const int ker_glb, Block* blk) const ;

  RangeRep* subker() const ;

  int volume() const ;

  int glb(const int sub, const int ker_glb) const ;

  void show() ;

protected :
  int numLayers, numProcs ;
  int f, f_s, k_step ;

private :
  void euler() ;

MEMORY_POOL_DECS
} ;


// -------------------------------------------------------------
// Things to do
//
//   Revert to `CyclicSubRangeRep' derived from `CyclicRangeRep'.

