
// Operations on irregular block distributed ranges.

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `IrregDist': irregular block-distribution parameters          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class IrregDist {
protected :
  int findBlock(const int tem) const {

    // Based on `findblock' routine in Global Arrays.

    int ker_glb, found = 0 ;
    int candidate = (int) (scale * tem) ;

    if(map [candidate] > tem) {

      // search downward

      ker_glb = candidate - 1 ;
      while(ker_glb >= 0) {
	found = (map [ker_glb] <= tem) ;
	if(found) break ;
	ker_glb-- ;
      }
    }
    else {

      // search upward

      ker_glb = candidate ;
      while(ker_glb < nBlocks - 1) {
	found = (map [ker_glb + 1] > tem);
	if(found) break;
	ker_glb++;
      }
    }

    if(!found) ker_glb = nBlocks - 1 ;

    return ker_glb ;
  }

  int nBlocks ;
  int *map ;
  int *blocks ;
  int blockSize ;

  float scale ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `IrregRangeRep': irregular block-distributed template range.  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class IrregRangeRep : public RangeRep, public IrregDist {
public :
  IrregRangeRep(RangeRep* _kernel, int* _blocks) ;
  IrregRangeRep(const int _extent, RangeRep* _kernel, int* _map) ;

  ~IrregRangeRep() ;

  RangeRep* shell() const ;

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  void location(Location* loc, const int glb) const ;

  void block(const int ker_glb, Block* blk) const ;

  int volume() const ;

  int glb(const int sub, const int ker_glb) const ;

  void show() ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `IrregSubRangeRep': block-distributed subrange.               //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class IrregSubRangeRep : public SubRangeRep, public IrregDist {
public :
  IrregSubRangeRep(const int _extent, const int _base, const int _stride,
		   RangeRep* _parent, RangeRep* _kernel, IrregDist* dist) ;

  RangeRep* shell() const ;

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  void location(Location* loc, const int glb) const ;

  void block(const int ker_glb, Block* blk) const ;

  RangeRep* subker() const ;

  int volume() const ;

  int glb(const int sub, const int ker_glb) const ;

  void show() ;

protected :

  int ker_lo, ker_hi ;

MEMORY_POOL_DECS
} ;


// -------------------------------------------------------------
// Things to do
//

