
// Ranges and strides, and subscripts and indexes.

#include "RangeRep.h"
#include "CollapsedRangeRep.h"
#include "DimRangeRep.h"
#include "BlockRangeRep.h"
#include "CyclicRangeRep.h"
#include "StepRangeRep.h"
#include "IrregRangeRep.h"

//#define MAX_LEVELS 2 

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Range': handle class for range representations.              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

//enum Layout {LAY_STANDARD, LAY_PACKED, LAY_ADAPT, LAY_EXTENDED} ;

struct Stride ;
//class Stride ;

#define DEF_LB  0
#define DEF_UB  -2

struct Trip {
  Trip() : lb(DEF_LB), ub(DEF_UB), str(1) {}
  Trip(const int _lb) : lb(_lb), ub(DEF_UB), str(1) {}
  Trip(const int _lb, const int _ub) : lb(_lb), ub(_ub), str(1) {}
  Trip(const int _lb, const int _ub, const int _str) :
      lb(_lb), ub(_ub), str(_str) {}

      // probably should check _ub >= -1.

  const int lb, ub, str ;
} ;

class Range {
public :
  // Constructors and destructors

  Range() : rep(0) {}

  Range(const Range& x) : rep(x.rep) {
    if(rep) rep->incr() ;
  }

//  Range(const int extent) {
//    rep = new CollapsedRangeRep(extent) ;
//    rep->incr() ;
//  }
//
//  Range(Dimension dim) {
//    rep = new DimRangeRep(dim) ;
//    rep->incr() ;
//  }

  ~Range() {if(rep) rep->decr() ; }

  // Subranges

  Range subrng(const int extent,
               const int base = 0, const int stride = 1) const {
    if(extent != 0) {
      const int parExtent = size() ;
      const int top = base + stride * (extent - 1) ;
      if(base < 0 || base >= parExtent || top < 0 || top >= parExtent)
        AdlibError("Range :: subrng: subrange out of range") ;
    }

    return Range(rep->subrng(extent, base, stride)) ;
  }

  Range triplet(const int base, const int top, const int stride = 1) const {
    int extent ;
    if((stride > 0 && top < base) || (stride < 0 && top > base))
      extent = 0 ;
    else {
      const int parExtent = size() ;
      if(base < 0 || base >= parExtent || top < 0 || top >= parExtent)
        AdlibError("Range :: trip: subrange out of range") ;
      else
        extent = (top - base) / stride + 1 ;
    }

    return Range(rep->subrng(extent, base, stride)) ;
  }

  // Inquiries

  int size() const {return rep->size() ; }

  Dimension dim() const {return rep->dim() ; }

  int lev() const {return rep->lev() ; }
  Range ker() const { return Range(rep->ker()) ; }
  Range shell() const { return Range(rep->shell()) ; }

  Range tem() const { return Range(rep->tem()) ; }
  int bas() const {return rep->bas() ; }
  int str() const {return rep->str() ; }

  Format format() const {return rep->format() ; }

  // Subscripting and iteration

  void location(Location* loc, const int glb) const {
    rep->location(loc, glb) ;
  }

  int idx(const Location& loc) const {
    return rep->idx(&loc) ;
  }

  void block(Block* blk, const int ker_glb) const {
    rep->block(ker_glb, blk) ;
  }

  int local(int* glb) const {return rep->local(glb) ; }

  Range subker() const {
    return Range(rep->subker()) ;
  }

  // `Stride' stuff

  int disp(const int sub) const {return rep->disp(sub) ; }
  int step(const int sub_stp) const {return rep->step(sub_stp) ; }

  int offset(const Location& i) const {return rep->offset(&i) ; }

  int volume() const {return rep->volume() ; }

  // Miscellaneous operators

  Location operator()(const int glb) {
    Location res;
    location(&res, glb) ;
    return res ;
  }

  Range operator()(const Trip& t) {
    const int base   = t.lb ;
    const int top    = t.ub != DEF_UB ? t.ub : size() - 1 ;
    const int stride = t.str ;
    return triplet(base, top, stride) ;
  }

  Range& operator=(const Range& x) {
    if(x.rep) x.rep->incr() ;
    if(rep)   rep->decr() ;
    rep = x.rep ;
    return *this ;
  }

  //operator RangeRep*() const {return rep ; }

  // Interim support for PCRC Fortran interface

  int glb(const int sub, const int ker_glb) const {
    return rep->glb(sub, ker_glb) ;
  }

  int& g_lb() {return rep->g_lb ; }
  int& g_stride() {return rep->g_stride ; }

  // debugging

  void show() {
    if(rep)
      rep->show() ;
    else
      AdlibPrintf("Range :: show: Null range\n") ;
  }
protected :
  Range(RangeRep* _rep) : rep(_rep) {
    if(rep) rep->incr() ;
  }

  RangeRep* rep ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CollapsedRange': collapsed range.                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class CollapsedRange : public Range {
public :
  CollapsedRange(const int extent) {
    rep = new CollapsedRangeRep(extent) ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `DimRange': range representing a process dimension.           //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class DimRange : public Range {
public :
  DimRange(Dimension dim) {
    rep = new DimRangeRep(dim) ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockRange': block-distributed range.                        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockRange : public Range {
public :
  BlockRange(const int extent, Dimension dim) {
    const int blockSize = (extent + dim.size() - 1) / dim.size() ;

    rep = new BlockRangeRep(extent, new DimRangeRep(dim), blockSize) ;
    rep->incr() ;
  }

  BlockRange(const int extent, Dimension dim, const int wlo, const int whi) {
    const int blockSize = (extent + dim.size() - 1) / dim.size() ;

    rep = new ExtendedBlockRangeRep(extent, new DimRangeRep(dim), blockSize,
                                    wlo, whi) ;
    rep->incr() ;
  }

  BlockRange(const int extent, Dimension dim,  const int blockSize) {
    rep = new BlockRangeRep(extent, new DimRangeRep(dim), blockSize) ;
    rep->incr() ;
  }

  BlockRange(const int extent, Dimension dim, const int blockSize,
	     const int wlo, const int whi) {
    rep = new ExtendedBlockRangeRep(extent, new DimRangeRep(dim), blockSize,
                                    wlo, whi) ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CyclicRange': cyclic-distributed range.                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class CyclicRange : public Range {
public :
  CyclicRange(const int extent, Dimension dim) {
    const int numCycles = (extent + dim.size() - 1) / dim.size() ;
    rep = new CyclicRangeRep(extent, new DimRangeRep(dim), numCycles) ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `BlockCyclicRange': block-cyclically distributed range.       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class BlockCyclicRange : public Range {
public :
  BlockCyclicRange(const int extent, Dimension dim, const int blockSize) {
    const int numBlocks = (extent + blockSize - 1) / blockSize ;
    const int numCycles = (numBlocks + dim.size() - 1) / dim.size() ;
    RangeRep* ker = new CyclicRangeRep(numBlocks, new DimRangeRep(dim),
                                       numCycles) ;
    rep = new BlockCyclicRangeRep(extent, ker, blockSize) ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `IrregRange': irregular block-distributed range.              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class IrregRange : public Range {
public :
  IrregRange(Dimension dim, int* blocks) {
    rep = new IrregRangeRep(new DimRangeRep(dim), blocks) ;
    rep->incr() ;
  }

  IrregRange(const int extent, Dimension dim, int* map) {
    rep = new IrregRangeRep(extent, new DimRangeRep(dim), map) ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `StepRange': stepped block-distributed range.                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class StepRange : public Range {
public :
  StepRange(const int extent, Dimension dim) {
    rep = new StepRangeRep(new DimRangeRep(dim),
                           extent / dim.size(), extent % dim.size()) ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ADAPTBlockRange': block-distributed range yielding arrays    //
//                           with special packed memory layout         //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ADAPTBlockRange : public Range {
public :
  ADAPTBlockRange(const int extent, const int base, const int stride,
                  const int parExtent, Dimension dim, const int blockSize) {
    BlockRangeRep* par = new BlockRangeRep(parExtent, new DimRangeRep(dim),
                                           blockSize) ;

    if(extent != par->size() || stride < 0)
      rep = new PackedBlockSubRangeRep(extent, base, stride,
				       par, par->ker(), blockSize,
				       stride > 0 ? stride : -stride) ;
    else
      rep = par ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ADAPTCyclicRange': cyclic-distributed range yielding arrays  //
//                           with special truncated memory layout.     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ADAPTCyclicRange : public Range {
public :
  ADAPTCyclicRange(const int extent, const int base, const int stride,
                   Dimension dim) {
    const int parExtent =
        (stride > 0 ? base + stride * (extent - 1) : base) + 1 ;
    const int numCycles = (parExtent + dim.size() - 1) / dim.size() ;

    RangeRep* par = new CyclicRangeRep(parExtent, new DimRangeRep(dim),
                                       numCycles) ;

    rep = (extent != par->size() || stride < 0) ?
              par->subrng(extent, base, stride) : par ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `ADAPTBlockCyclicRange': block-cyclically distributed range,  //
//                                yielding arrays with special packed  //
//                                memory layout.                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class ADAPTBlockCyclicRange : public Range {
public :
  ADAPTBlockCyclicRange(const int extent, const int base, const int stride,
                        Dimension dim, const int blockSize) {
    const int numBlocks =
        (stride > 0 ? base + stride * (extent - 1) : base) / blockSize + 1 ;
    const int numCycles = (numBlocks + dim.size() - 1) / dim.size() ;

    RangeRep* ker = new CyclicRangeRep(numBlocks, new DimRangeRep(dim),
                                       numCycles) ;
    BlockRangeRep* par = new BlockCyclicRangeRep(numBlocks * blockSize,
                                                 ker, blockSize) ;

    if(extent != par->size() || stride < 0)
      rep = new PackedBlockCyclicSubRangeRep(extent, base, stride,
				             par, par->ker(), blockSize,
				             stride > 0 ? stride : -stride) ;
    else
      rep = par ;
    rep->incr() ;
  }
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Stride'                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

struct Stride {
public :
  Stride() : itemLen(0) {}

  Stride(Range _parent, const int _itemLen) :
               parent(_parent), itemLen(_itemLen) {}

  Stride ker() const {
    return Stride(parent.ker(), parent.shell().volume() * itemLen) ;
  }

  Stride shell() const {
    return Stride(parent.shell(), itemLen) ;
  }

  int len() const {return itemLen ; }

  int offset(const Location& i) const {
    return itemLen * parent.offset(i) ;
  }

  int disp(const int sub) const {
    return itemLen * parent.disp(sub) ;
  }

  int step(const int sub_stp) const {
    return itemLen * parent.step(sub_stp) ;
  }

  Range parent ;
  int itemLen ;
} ;


// Iterator (index) classes

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `LocBlocksIndex': enumerate local blocks of range.            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class LocBlocksIndex : public Location, public Block {
public :
  LocBlocksIndex(Range _x) : x(_x), level(_x.lev()) {

    switch(level) {
      case 0 :
        d = x ;

        break ;

      case 1 :
        d = x.subker() ;

        break ;

      case 2 :
        k = x.subker() ;
        d = k.subker() ;

        break ;
    }

    dim = d.dim() ;
  }

  void beginLocBlk() {
    int d_glb ;
    if(d.local(&d_glb)) {
      crd = dim.crd() ;

      switch(level) {
        case 0 :
          glb_bas = d_glb ;
          glb_stp = 1 ;  // Feeds into subrange constructors: mustn't be zero.
	  tem_bas = crd ;
          tem_stp = 0 ;
	  sub_bas = 0 ;
          sub_stp = 0 ;
          count   = 1 ;

          looping = 1 ;

          tem     = crd ;
          sub     = 0   ;  // Used in `Index :: next()'.
          break ;

        case 1 :
          x.block(this, crd) ;
          looping = count ;

          tem = tem_bas ;
          sub = sub_bas ;
          break ;

        case 2 : {
          looping = 1 ;

          k.block(&k_block, crd) ;
          k_tem = k_block.tem_bas ;
          blk   = k_block.sub_bas ;
          k_l = 0 ;
          scan() ;

          tem = tem_bas ;
          sub = sub_bas ;
          break ;
        }
      }
    }
    else
      looping = 0 ;
  }

  int test() {return looping ; }

  void nextLocBlk() {
    if(level < 2)
      looping = 0 ;
    else {
      blk   += k_block.sub_stp ;
      k_tem += k_block.tem_stp ;

      k_l++ ;
      scan() ;

      tem = tem_bas ;
      sub = sub_bas ;
    }
  }

private :

  void scan() {

    // Starting from current position in kernel range, `k',
    // find first non-empty block.

    // (Only used for level = 2.)

    for(;;)
      if(k_l < k_block.count) {
        x.block(this, k_tem) ;

        if(count)
          break ;
        else {
          k_tem += k_block.tem_stp ;
          blk   += k_block.sub_stp ;

          k_l++ ;
        }
      }
      else {
        looping = 0 ;
        break ;
      }
  }

protected :
  Range x, d ;
  int level ;

  int looping ;

  // For level 2 ranges:

  Range k ;

  Block k_block ;
  int k_l, k_tem ;

private :
  LocBlocksIndex(const LocBlocksIndex&) {} ;  // No copying allowed.
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `AllBlocksIndex': enumerate all blocks of range.              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class AllBlocksIndex : public LocBlocksIndex {
public :
  AllBlocksIndex(Range _x) : LocBlocksIndex(_x) {}

  void beginAllBlk() {
    crd = d.bas() ;
    switch(level) {
      case 0 :
	glb_bas = 0 ;
        glb_stp = 1 ;  // Feeds into subrange constructors: mustn't be zero.
	tem_bas = crd ;
        tem_stp = 0 ;
	sub_bas = 0 ;
        sub_stp = 0 ;
        count   = 1 ;

	looping = d.size() ;

        tem     = crd ;
	break ;

      case 1 :
	d_l     = 0 ;

	looping = d.size() ;
	if(looping) {
	  x.block(this, crd) ;
	  while(count == 0) {
	    crd += d.str() ;

	    d_l++ ;
	    if(d_l == d.size()) {
	      looping = 0 ;
	      break ;
	    }
	    else
	      x.block(this, crd) ;
	  }
          tem = tem_bas ;
	  sub = sub_bas ;
	}
	break ;

      case 2 : {
	d_l     = 0 ;

        looping = 1 ;

	k.block(&k_block, crd) ;
        k_tem = k_block.tem_bas ;
        blk   = k_block.sub_bas ;

	k_l = 0 ;
        scan() ;

	tem = tem_bas ;
	sub = sub_bas ;
	break ;
      }
    }
  }

  void nextAllBlk() {
    switch(level) {
      case 0 :
	crd += d.str() ;
	glb_bas++ ;

	looping = glb_bas < d.size() ;

        tem = crd ;
	break ;

      case 1 :
        do {
	  crd += d.str() ;

	  d_l++ ;

	  if(d_l == d.size()) {
	    looping = 0 ;
	    break ;
	  }
	  else
	    x.block(this, crd) ;

        } while (count == 0) ;

	tem = tem_bas ;
	sub = sub_bas ;
	break ;

      case 2 :
        blk   += k_block.sub_stp ;
        k_tem += k_block.tem_stp ;

        k_l++ ;
        scan() ;

	tem = tem_bas ;
	sub = sub_bas ;
	break ;
    }
  }

private :

  void scan() {

    // Starting from current position in kernel range, `k',
    // find first non-empty block.

    // (Only used for level = 2.)

    for(;;)
      if(k_l < k_block.count) {
        x.block(this, k_tem) ;

        if(count)
          break ;
        else {
          k_tem += k_block.tem_stp ;
          blk   += k_block.sub_stp ;

          k_l++ ;
        }
      }
      else {
        crd += d.str() ;

        d_l++ ;
        if(d_l == d.size()) {
          looping = 0 ;
          break ;
        }
        else {
          k.block(&k_block, crd) ;
          k_tem = k_block.tem_bas ;
          blk   = k_block.sub_bas ;

          k_l = 0 ;
        }
      }
  }

private :

  int d_l ;
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `Index'.                                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class Index : public LocBlocksIndex {
public :
  Index(Range _x) : LocBlocksIndex(_x) {}

  void begin() {
    beginLocBlk() ;

    l   = 0 ;
  }

  void next() {
    tem += tem_stp ;
    sub += sub_stp ;

    l++ ;
    if(l == count) {
      nextLocBlk() ;

      l   = 0 ;
    }
  }

private :
  int l ;

} ;




// ---------------------------------------------------------------
//
// Things to do:
//
//   Restrict definiton of `Range :: disp' to collapsed ranges?

