
// Operations on stepped block distributed ranges.

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `StepRangeRep': stepped block-distributed template range.     //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class StepRangeRep : public RangeRep {
public :
  StepRangeRep(RangeRep* _kernel, const int _blockSize, const int _stepCrd) :
          RangeRep(_blockSize * _kernel->size() + _stepCrd,
                   _kernel, _kernel->lev() + 1, _kernel->dim(), DIST_STEP),
          blockSize(_blockSize), stepCrd(_stepCrd),
          stepSub(_stepCrd * (_blockSize + 1)) {
    kernel->incr() ;
  }

  ~StepRangeRep() {
    kernel->decr() ;
  }

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  RangeRep* shell() const ;

  void location(Location* loc, const int glb) const ;

  void block(const int ker_glb, Block* blk) const ;

  int volume() const ;

  int glb(const int sub, const int ker_glb) const ;

  void show() ;

protected :

  int blockSize, stepCrd, stepSub ;

MEMORY_POOL_DECS
} ;


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `StepSubRangeRep': block-distributed subrange.                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

class StepSubRangeRep : public SubRangeRep {
public :
  StepSubRangeRep(const int _extent,
		  const int _base, const int _stride,
		  RangeRep* _parent,
		  RangeRep* _kernel,
                  const int _blockSize, const int _stepCrd) :
          SubRangeRep(_extent, _base, _stride, _parent,
                      _kernel, _kernel->lev() + 1, _kernel->dim(), DIST_STEP),
          blockSize(_blockSize), stepCrd(_stepCrd),
          stepSub(_stepCrd * (_blockSize + 1)) {

    if(stride > 0) {
      ker_lo = to_ker_glb(base) ;
      ker_hi = to_ker_glb(base + stride * (extent - 1)) ;
    }
    else {
      ker_hi = to_ker_glb(base) ;
      ker_lo = to_ker_glb(base + stride * (extent - 1)) ;
    }
  }

  RangeRep* subrng(const int _extent, const int _base, const int _stride) ;

  RangeRep* shell() const ;

  void location(Location* loc, const int glb) const ;

  void block(const int ker_glb, Block* blk) const ;

  RangeRep* subker() const ;

  int glb(const int sub, const int ker_glb) const ;

  int volume() const ;

  void show() ;

protected :

  int to_ker_glb(const int tem) const {
    if(tem < stepSub)
      return tem / (blockSize + 1) ;
    else
      return (tem - stepCrd) / blockSize ;
  }

  int blockSize, stepCrd, stepSub ;

  int ker_lo, ker_hi ;

MEMORY_POOL_DECS
} ;


// -------------------------------------------------------------
// Things to do
//

