
// Array transformational functions.

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `remap', `shift', `skew', `multishift' and `writeHalo'.             //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
inline T broadcast(const Section0<T>& src) {
  Array0<T> dst ;
  Remap schedule(&dst, &src, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;

  return dst() ;
}

template<class T>
inline void remap(const Section0<T>& dst, const Section0<T>& src) {
  Remap schedule(&dst, &src, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void remap(const Section1<T>& dst, const Section1<T>& src) {
  Remap schedule(&dst, &src, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void remap(const Section2<T>& dst, const Section2<T>& src) {
  Remap schedule(&dst, &src, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void remap(const Section3<T>& dst, const Section3<T>& src) {
  Remap schedule(&dst, &src, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void shift(const Section1<T>& dst, const Section1<T>& src,
                  const int shift, const int dim, const Mode mode) {
  Shift schedule(&dst, &src, sizeof(T), shift, dim, mode) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void shift(const Section2<T>& dst, const Section2<T>& src,
                  const int shift, const int dim, const Mode mode) {
  Shift schedule(&dst, &src, sizeof(T), shift, dim, mode) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void shift(const Section3<T>& dst, const Section3<T>& src,
                  const int shift, const int dim, const Mode mode) {
  Shift schedule(&dst, &src, sizeof(T), shift, dim, mode) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void skew(const Section2<T>& dst, const Section2<T>& src,
                 const Section1<int>& shf,
                 const int dim, const Mode mode) {
  Skew schedule(&dst, &src, sizeof(T), &shf, shf.dat(), dim, mode) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void skew(const Section3<T>& dst, const Section3<T>& src,
                 const Section2<int>& shf,
                 const int dim, const Mode mode) {
  Skew schedule(&dst, &src, sizeof(T), &shf, shf.dat(), dim, mode) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void multiShift(const Section1<T>& dst, const Section1<T>& src,
                       int* shifts, const Mode* modes) {
  MultiShift schedule(&dst, &src, sizeof(T), shifts, modes) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void multiShift(const Section2<T>& dst, const Section2<T>& src,
                       int* shifts, const Mode* modes) {
  MultiShift schedule(&dst, &src, sizeof(T), shifts, modes) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void multiShift(const Section3<T>& dst, const Section3<T>& src,
                       int* shifts, const Mode* modes) {
  MultiShift schedule(&dst, &src, sizeof(T), shifts, modes) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void writeHalo(const Section1<T>& src,
                      const int* wlo, const int* whi, const Mode* modes = 0) {
  WriteHalo schedule(&src, sizeof(T), wlo, whi, modes) ;

  schedule.execute(src.dat()) ;
}

template<class T>
inline void writeHalo(const Section2<T>& src,
                      const int* wlo, const int* whi, const Mode* modes = 0) {
  WriteHalo schedule(&src, sizeof(T), wlo, whi, modes) ;

  schedule.execute(src.dat()) ;
}

template<class T>
inline void writeHalo(const Section3<T>& src,
                      const int* wlo, const int* whi, const Mode* modes = 0) {
  WriteHalo schedule(&src, sizeof(T), wlo, whi, modes) ;

  schedule.execute(src.dat()) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `gather', `scatter', `vecGather', `vecScatter' and `reshape'.       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

// Simple gather

template<class T>
inline void gather(const Section1<T>& b, const Section1<T>& a,
                   const Section1<int>& s_1) {
  Gather schedule(&b, &a, sizeof(T),
                  Seq1<const DAD*>(&s_1), Seq1<const int*>(s_1.dat()),
                  0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void gather(const Section2<T>& b, const Section1<T>& a,
                   const Section2<int>& s_1) {
  Gather schedule(&b, &a, sizeof(T),
                  Seq1<const DAD*>(&s_1), Seq1<const int*>(s_1.dat()),
                  0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void gather(const Section3<T>& b, const Section1<T>& a,
                   const Section3<int>& s_1) {
  Gather schedule(&b, &a, sizeof(T),
                  Seq1<const DAD*>(&s_1), Seq1<const int*>(s_1.dat()),
                  0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void gather(const Section1<T>& b, const Section2<T>& a,
                   const Section1<int>& s_1, const Section1<int>& s_2) {
  Gather schedule(&b, &a, sizeof(T),
                  Seq2<const DAD*>(&s_1, &s_2),
                  Seq2<const int*>(s_1.dat(), s_2.dat()),
                  0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void gather(const Section2<T>& b, const Section2<T>& a,
                   const Section2<int>& s_1, const Section2<int>& s_2) {
  Gather schedule(&b, &a, sizeof(T),
                  Seq2<const DAD*>(&s_1, &s_2),
                  Seq2<const int*>(s_1.dat(), s_2.dat()),
                  0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void gather(const Section3<T>& b, const Section2<T>& a,
                   const Section3<int>& s_1, const Section3<int>& s_2) {
  Gather schedule(&b, &a, sizeof(T),
                  Seq2<const DAD*>(&s_1, &s_2),
                  Seq2<const int*>(s_1.dat(), s_2.dat()),
                  0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void gather(const Section1<T>& b, const Section3<T>& a,
                   const Section1<int>& s_1, const Section1<int>& s_2,
                       const Section1<int>& s_3) {
  Gather schedule(&b, &a, sizeof(T),
                  Seq3<const DAD*>(&s_1, &s_2, &s_3),
                  Seq3<const int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                  0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void gather(const Section2<T>& b, const Section3<T>& a,
                   const Section2<int>& s_1, const Section2<int>& s_2,
                       const Section2<int>& s_3) {
  Gather schedule(&b, &a, sizeof(T),
                  Seq3<const DAD*>(&s_1, &s_2, &s_3),
                  Seq3<const int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                  0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void gather(const Section3<T>& b, const Section3<T>& a,
                   const Section3<int>& s_1, const Section3<int>& s_2,
                       const Section3<int>& s_3) {
  Gather schedule(&b, &a, sizeof(T),
                  Seq3<const DAD*>(&s_1, &s_2, &s_3),
                  Seq3<const int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                  0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}


// Masked gather

template<class T>
inline void gather(const Section1<T>& b, const Section1<T>& a,
                   const Section1<int>& s_1,
                   const Section1<int>& msk) {
  Gather schedule(&b, &a, sizeof(T),
                  Seq1<const DAD*>(&s_1), Seq1<const int*>(s_1.dat()),
                  &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void gather(const Section2<T>& b, const Section1<T>& a,
                   const Section2<int>& s_1,
                   const Section2<int>& msk) {
  Gather schedule(&b, &a, sizeof(T),
                  Seq1<const DAD*>(&s_1), Seq1<const int*>(s_1.dat()),
                  &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void gather(const Section3<T>& b, const Section1<T>& a,
                   const Section3<int>& s_1,
                   const Section3<int>& msk) {
  Gather schedule(&b, &a, sizeof(T),
                  Seq1<const DAD*>(&s_1), Seq1<const int*>(s_1.dat()),
                  &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void gather(const Section1<T>& b, const Section2<T>& a,
                   const Section1<int>& s_1, const Section1<int>& s_2,
                   const Section1<int>& msk) {
  Gather schedule(&b, &a, sizeof(T),
                  Seq2<const DAD*>(&s_1, &s_2),
                  Seq2<const int*>(s_1.dat(), s_2.dat()),
                  &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void gather(const Section2<T>& b, const Section2<T>& a,
                   const Section2<int>& s_1, const Section2<int>& s_2,
                   const Section2<int>& msk) {
  Gather schedule(&b, &a, sizeof(T),
                  Seq2<const DAD*>(&s_1, &s_2),
                  Seq2<const int*>(s_1.dat(), s_2.dat()),
                  &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void gather(const Section3<T>& b, const Section2<T>& a,
                   const Section3<int>& s_1, const Section3<int>& s_2,
                   const Section3<int>& msk) {
  Gather schedule(&b, &a, sizeof(T),
                  Seq2<const DAD*>(&s_1, &s_2),
                  Seq2<const int*>(s_1.dat(), s_2.dat()),
                  &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void gather(const Section1<T>& b, const Section3<T>& a,
                   const Section1<int>& s_1, const Section1<int>& s_2,
                       const Section1<int>& s_3,
                   const Section1<int>& msk) {
  Gather schedule(&b, &a, sizeof(T),
                  Seq3<const DAD*>(&s_1, &s_2, &s_3),
                  Seq3<const int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                  &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void gather(const Section2<T>& b, const Section3<T>& a,
                   const Section2<int>& s_1, const Section2<int>& s_2,
                       const Section2<int>& s_3,
                   const Section2<int>& msk) {
  Gather schedule(&b, &a, sizeof(T),
                  Seq3<const DAD*>(&s_1, &s_2, &s_3),
                  Seq3<const int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                  &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void gather(const Section3<T>& b, const Section3<T>& a,
                   const Section3<int>& s_1, const Section3<int>& s_2,
                       const Section3<int>& s_3,
                   const Section3<int>& msk) {
  Gather schedule(&b, &a, sizeof(T),
                  Seq3<const DAD*>(&s_1, &s_2, &s_3),
                  Seq3<const int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                  &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}


// Simple scatter

template<class T>
inline void scatter(const Section1<T>& b, const Section1<T>& a,
                    const Section1<int>& s_1) {
  Scatter schedule(&b, &a, sizeof(T),
                   Seq1<const DAD*>(&s_1), Seq1<const int*>(s_1.dat()),
                   0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatter(const Section2<T>& b, const Section1<T>& a,
                    const Section2<int>& s_1) {
  Scatter schedule(&b, &a, sizeof(T),
                   Seq1<const DAD*>(&s_1), Seq1<const int*>(s_1.dat()),
                   0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatter(const Section3<T>& b, const Section1<T>& a,
                    const Section3<int>& s_1) {
  Scatter schedule(&b, &a, sizeof(T),
                   Seq1<const DAD*>(&s_1), Seq1<const int*>(s_1.dat()),
                   0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatter(const Section1<T>& b, const Section2<T>& a,
                    const Section1<int>& s_1, const Section1<int>& s_2) {
  Scatter schedule(&b, &a, sizeof(T),
                   Seq2<const DAD*>(&s_1, &s_2),
                   Seq2<const int*>(s_1.dat(), s_2.dat()),
                   0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatter(const Section2<T>& b, const Section2<T>& a,
                    const Section2<int>& s_1, const Section2<int>& s_2) {
  Scatter schedule(&b, &a, sizeof(T),
                   Seq2<const DAD*>(&s_1, &s_2),
                   Seq2<const int*>(s_1.dat(), s_2.dat()),
                   0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatter(const Section3<T>& b, const Section2<T>& a,
                    const Section3<int>& s_1, const Section3<int>& s_2) {
  Scatter schedule(&b, &a, sizeof(T),
                   Seq2<const DAD*>(&s_1, &s_2),
                   Seq2<const int*>(s_1.dat(), s_2.dat()),
                   0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatter(const Section1<T>& b, const Section3<T>& a,
                    const Section1<int>& s_1, const Section1<int>& s_2,
                    const Section1<int>& s_3) {
  Scatter schedule(&b, &a, sizeof(T),
                   Seq3<const DAD*>(&s_1, &s_2, &s_3),
                   Seq3<const int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                   0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatter(const Section2<T>& b, const Section3<T>& a,
                    const Section2<int>& s_1, const Section2<int>& s_2,
                    const Section2<int>& s_3) {
  Scatter schedule(&b, &a, sizeof(T),
                   Seq3<const DAD*>(&s_1, &s_2, &s_3),
                   Seq3<const int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                   0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatter(const Section3<T>& b, const Section3<T>& a,
                    const Section3<int>& s_1, const Section3<int>& s_2,
                    const Section3<int>& s_3) {
  Scatter schedule(&b, &a, sizeof(T),
                   Seq3<const DAD*>(&s_1, &s_2, &s_3),
                   Seq3<const int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                   0, 0) ;

  schedule.execute(b.dat(), a.dat()) ;
}


// Masked scatter

template<class T>
inline void scatter(const Section1<T>& b, const Section1<T>& a,
                    const Section1<int>& s_1,
                    const Section1<int>& msk) {
  Scatter schedule(&b, &a, sizeof(T),
                   Seq1<const DAD*>(&s_1), Seq1<const int*>(s_1.dat()),
                   &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatter(const Section2<T>& b, const Section1<T>& a,
                    const Section2<int>& s_1,
                    const Section2<int>& msk) {
  Scatter schedule(&b, &a, sizeof(T),
                   Seq1<const DAD*>(&s_1), Seq1<const int*>(s_1.dat()),
                   &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatter(const Section3<T>& b, const Section1<T>& a,
                    const Section3<int>& s_1,
                    const Section3<int>& msk) {
  Scatter schedule(&b, &a, sizeof(T),
                   Seq1<const DAD*>(&s_1), Seq1<const int*>(s_1.dat()),
                   &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatter(const Section1<T>& b, const Section2<T>& a,
                    const Section1<int>& s_1, const Section1<int>& s_2,
                    const Section1<int>& msk) {
  Scatter schedule(&b, &a, sizeof(T),
                   Seq2<const DAD*>(&s_1, &s_2),
                   Seq2<const int*>(s_1.dat(), s_2.dat()),
                   &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatter(const Section2<T>& b, const Section2<T>& a,
                    const Section2<int>& s_1, const Section2<int>& s_2,
                    const Section2<int>& msk) {
  Scatter schedule(&b, &a, sizeof(T),
                   Seq2<const DAD*>(&s_1, &s_2),
                   Seq2<const int*>(s_1.dat(), s_2.dat()),
                   &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatter(const Section3<T>& b, const Section2<T>& a,
                    const Section3<int>& s_1, const Section3<int>& s_2,
                    const Section3<int>& msk) {
  Scatter schedule(&b, &a, sizeof(T),
                   Seq2<const DAD*>(&s_1, &s_2),
                   Seq2<const int*>(s_1.dat(), s_2.dat()),
                   &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatter(const Section1<T>& b, const Section3<T>& a,
                    const Section1<int>& s_1, const Section1<int>& s_2,
                    const Section1<int>& s_3,
                    const Section1<int>& msk) {
  Scatter schedule(&b, &a, sizeof(T),
                   Seq3<const DAD*>(&s_1, &s_2, &s_3),
                   Seq3<const int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                   &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatter(const Section2<T>& b, const Section3<T>& a,
                    const Section2<int>& s_1, const Section2<int>& s_2,
                    const Section2<int>& s_3,
                    const Section2<int>& msk) {
  Scatter schedule(&b, &a, sizeof(T),
                   Seq3<const DAD*>(&s_1, &s_2, &s_3),
                   Seq3<const int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                   &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatter(const Section3<T>& b, const Section3<T>& a,
                    const Section3<int>& s_1, const Section3<int>& s_2,
                    const Section3<int>& s_3,
                    const Section3<int>& msk) {
  Scatter schedule(&b, &a, sizeof(T),
                   Seq3<const DAD*>(&s_1, &s_2, &s_3),
                   Seq3<const int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                   &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}


// Combining scatter

template<class T>
inline void scatterComb(const Section1<T>& b, const Section1<T>& a,
                        const Section1<int>& s_1,
                        void (*op)(void*, void*)) {
  ScatterComb schedule(&b, &a, sizeof(T),
                       Seq1<const DAD*>(&s_1), Seq1<const int*>(s_1.dat()),
                       0, 0,
                       op) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatterComb(const Section2<T>& b, const Section1<T>& a,
                        const Section2<int>& s_1,
                        void (*op)(void*, void*)) {
  ScatterComb schedule(&b, &a, sizeof(T),
                       Seq1<const DAD*>(&s_1), Seq1<const int*>(s_1.dat()),
                       0, 0,
                       op) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatterComb(const Section3<T>& b, const Section1<T>& a,
                        const Section3<int>& s_1,
                        void (*op)(void*, void*)) {
  ScatterComb schedule(&b, &a, sizeof(T),
                       Seq1<const DAD*>(&s_1), Seq1<const int*>(s_1.dat()),
                       0, 0,
                       op) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatterComb(const Section1<T>& b, const Section2<T>& a,
                        const Section1<int>& s_1,
                        const Section1<int>& s_2,
                        void (*op)(void*, void*)) {
  ScatterComb schedule(&b, &a, sizeof(T),
                       Seq2<const DAD*>(&s_1, &s_2),
                       Seq2<const int*>(s_1.dat(), s_2.dat()),
                       0, 0,
                       op) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatterComb(const Section2<T>& b, const Section2<T>& a,
                        const Section2<int>& s_1,
                        const Section2<int>& s_2,
                        void (*op)(void*, void*)) {
  ScatterComb schedule(&b, &a, sizeof(T),
                       Seq2<const DAD*>(&s_1, &s_2),
                       Seq2<const int*>(s_1.dat(), s_2.dat()),
                       0, 0,
                       op) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatterComb(const Section3<T>& b, const Section2<T>& a,
                        const Section3<int>& s_1,
                        const Section3<int>& s_2,
                        void (*op)(void*, void*)) {
  ScatterComb schedule(&b, &a, sizeof(T),
                       Seq2<const DAD*>(&s_1, &s_2),
                       Seq2<const int*>(s_1.dat(), s_2.dat()),
                       0, 0,
                       op) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatterComb(const Section1<T>& b, const Section3<T>& a,
                        const Section1<int>& s_1,
                        const Section1<int>& s_2,
                        const Section1<int>& s_3,
                        void (*op)(void*, void*)) {
  ScatterComb schedule(&b, &a, sizeof(T),
                       Seq3<const DAD*>(&s_1, &s_2, &s_3),
                       Seq3<const int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                       0, 0,
                       op) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatterComb(const Section2<T>& b, const Section3<T>& a,
                        const Section2<int>& s_1,
                        const Section2<int>& s_2,
                        const Section2<int>& s_3,
                        void (*op)(void*, void*)) {
  ScatterComb schedule(&b, &a, sizeof(T),
                       Seq3<const DAD*>(&s_1, &s_2, &s_3),
                       Seq3<const int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                       0, 0,
                       op) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatterComb(const Section3<T>& b, const Section3<T>& a,
                        const Section3<int>& s_1,
                        const Section3<int>& s_2,
                        const Section3<int>& s_3,
                        void (*op)(void*, void*)) {
  ScatterComb schedule(&b, &a, sizeof(T),
                       Seq3<const DAD*>(&s_1, &s_2, &s_3),
                       Seq3<const int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                       0, 0,
                       op) ;

  schedule.execute(b.dat(), a.dat()) ;
}


// Masked combining scatter

template<class T>
inline void scatterComb(const Section1<T>& b, const Section1<T>& a,
                        const Section1<int>& s_1,
                        void (*op)(void*, void*),
                        const Section1<int>& msk) {
  ScatterComb schedule(&b, &a, sizeof(T),
                       Seq1<const DAD*>(&s_1), Seq1<const int*>(s_1.dat()),
                       &msk, msk.dat(),
                       op) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatterComb(const Section2<T>& b, const Section1<T>& a,
                        const Section2<int>& s_1,
                        void (*op)(void*, void*),
                        const Section2<int>& msk) {
  ScatterComb schedule(&b, &a, sizeof(T),
                       Seq1<const DAD*>(&s_1), Seq1<const int*>(s_1.dat()),
                       &msk, msk.dat(),
                       op) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatterComb(const Section3<T>& b, const Section1<T>& a,
                        const Section3<int>& s_1,
                        void (*op)(void*, void*),
                        const Section3<int>& msk) {
  ScatterComb schedule(&b, &a, sizeof(T),
                       Seq1<const DAD*>(&s_1), Seq1<const int*>(s_1.dat()),
                       &msk, msk.dat(),
                       op) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatterComb(const Section1<T>& b, const Section2<T>& a,
                        const Section1<int>& s_1,
                        const Section1<int>& s_2,
                        void (*op)(void*, void*),
                        const Section1<int>& msk) {
  ScatterComb schedule(&b, &a, sizeof(T),
                       Seq2<const DAD*>(&s_1, &s_2),
                       Seq2<const int*>(s_1.dat(), s_2.dat()),
                       &msk, msk.dat(),
                       op) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatterComb(const Section2<T>& b, const Section2<T>& a,
                        const Section2<int>& s_1,
                        const Section2<int>& s_2,
                        void (*op)(void*, void*),
                        const Section2<int>& msk) {
  ScatterComb schedule(&b, &a, sizeof(T),
                       Seq2<const DAD*>(&s_1, &s_2),
                       Seq2<const int*>(s_1.dat(), s_2.dat()),
                       &msk, msk.dat(),
                       op) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatterComb(const Section3<T>& b, const Section2<T>& a,
                        const Section3<int>& s_1,
                        const Section3<int>& s_2,
                        void (*op)(void*, void*),
                        const Section3<int>& msk) {
  ScatterComb schedule(&b, &a, sizeof(T),
                       Seq2<const DAD*>(&s_1, &s_2),
                       Seq2<const int*>(s_1.dat(), s_2.dat()),
                       &msk, msk.dat(),
                       op) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatterComb(const Section1<T>& b, const Section3<T>& a,
                        const Section1<int>& s_1,
                        const Section1<int>& s_2,
                        const Section1<int>& s_3,
                        void (*op)(void*, void*),
                        const Section1<int>& msk) {
  ScatterComb schedule(&b, &a, sizeof(T),
                       Seq3<const DAD*>(s_1, s_2, s_3),
                       Seq3<const int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                       &msk, msk.dat(),
                       op) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatterComb(const Section2<T>& b, const Section3<T>& a,
                        const Section2<int>& s_1,
                        const Section2<int>& s_2,
                        const Section2<int>& s_3,
                        void (*op)(void*, void*),
                        const Section2<int>& msk) {
  ScatterComb schedule(&b, &a, sizeof(T),
                       Seq3<const DAD*>(s_1, s_2, s_3),
                       Seq3<const int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                       &msk, msk.dat(),
                       op) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void scatterComb(const Section3<T>& b, const Section3<T>& a,
                        const Section3<int>& s_1,
                        const Section3<int>& s_2,
                        const Section3<int>& s_3,
                        void (*op)(void*, void*),
                        const Section3<int>& msk) {
  ScatterComb schedule(&b, &a, sizeof(T),
                       Seq3<const DAD*>(s_1, s_2, s_3),
                       Seq3<const int*>(s_1.dat(), s_2.dat(), s_3.dat()),
                       &msk, msk.dat(),
                       op) ;

  schedule.execute(b.dat(), a.dat()) ;
}


// Vector subscript gather.

template<class T>
inline void vecGather(const Section1<T>& b, const Section1<T>& a,
                      const Section1<int>* s_1) {
  VecGather schedule(&b, &a, sizeof(T),
                     Seq1<const DAD*>(s_1),
                     Seq1<const int*>(s_1 ? s_1->dat() : 0)) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void vecGather(const Section2<T>& b, const Section2<T>& a,
                      const Section1<int>* s_1, const Section1<int>* s_2) {
  VecGather schedule(&b, &a, sizeof(T),
                     Seq2<const DAD*>(s_1, s_2),
                     Seq2<const int*>(s_1 ? s_1->dat() : 0,
                                      s_2 ? s_2->dat() : 0)) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void vecGather(const Section3<T>& b, const Section3<T>& a,
                      const Section1<int>* s_1, const Section1<int>* s_2,
                          const Section1<int>* s_3) {
  VecGather schedule(&b, &a, sizeof(T),
                     Seq3<const DAD*>(s_1, s_2, s_3),
                     Seq3<const int*>(s_1 ? s_1->dat() : 0,
                                      s_2 ? s_2->dat() : 0,
                                      s_3 ? s_3->dat() : 0)) ;

  schedule.execute(b.dat(), a.dat()) ;
}

// Masked vector subscript scatter

template<class T>
inline void vecGather(const Section1<T>& b, const Section1<T>& a,
                      const Section1<int>* s_1,
                      const Section1<int>& msk) {
  VecGatherMsk schedule(&b, &a, sizeof(T),
                        Seq1<const DAD*>(s_1),
                        Seq1<const int*>(s_1 ? s_1->dat() : 0),
                        &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void vecGather(const Section2<T>& b, const Section2<T>& a,
                      const Section1<int>* s_1, const Section1<int>* s_2,
                      const Section2<int>& msk) {
  VecGatherMsk schedule(&b, &a, sizeof(T),
                        Seq2<const DAD*>(s_1, s_2),
                        Seq2<const int*>(s_1 ? s_1->dat() : 0,
                                         s_2 ? s_2->dat() : 0),
                        &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void vecGather(const Section3<T>& b, const Section3<T>& a,
                      const Section1<int>* s_1, const Section1<int>* s_2,
                      const Section1<int>* s_3,
                      const Section3<int>& msk) {
  VecGatherMsk schedule(&b, &a, sizeof(T),
                        Seq3<const DAD*>(s_1, s_2, s_3),
                        Seq3<const int*>(s_1 ? s_1->dat() : 0,
                                         s_2 ? s_2->dat() : 0,
                                         s_3 ? s_3->dat() : 0),
                        &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

// Vector subscript scatter.

template<class T>
inline void vecScatter(const Section1<T>& b, const Section1<T>& a,
                       const Section1<int>* s_1) {
  VecScatter schedule(&b, &a, sizeof(T),
                      Seq1<const DAD*>(s_1),
                      Seq1<const int*>(s_1 ? s_1->dat() : 0)) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void vecScatter(const Section2<T>& b, const Section2<T>& a,
                       const Section1<int>* s_1, const Section1<int>* s_2) {
  VecScatter schedule(&b, &a, sizeof(T),
                      Seq2<const DAD*>(s_1, s_2),
                      Seq2<const int*>(s_1 ? s_1->dat() : 0,
                                       s_2 ? s_2->dat() : 0)) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void vecScatter(const Section3<T>& b, const Section3<T>& a,
                       const Section1<int>* s_1, const Section1<int>* s_2,
                           const Section1<int>* s_3) {
  VecScatter schedule(&b, &a, sizeof(T),
                      Seq3<const DAD*>(s_1, s_2, s_3),
                      Seq3<const int*>(s_1 ? s_1->dat() : 0,
                                       s_2 ? s_2->dat() : 0,
                                       s_3 ? s_3->dat() : 0)) ;

  schedule.execute(b.dat(), a.dat()) ;
}

// Masked vector subscript scatter

template<class T>
inline void vecScatter(const Section1<T>& b, const Section1<T>& a,
                       const Section1<int>* s_1,
                       const Section1<int>& msk) {
  VecScatterMsk schedule(&b, &a, sizeof(T),
                         Seq1<const DAD*>(s_1),
                         Seq1<const int*>(s_1 ? s_1->dat() : 0),
                         &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void vecScatter(const Section2<T>& b, const Section2<T>& a,
                       const Section1<int>* s_1, const Section1<int>* s_2,
                       const Section2<int>& msk) {
  VecScatterMsk schedule(&b, &a, sizeof(T),
                         Seq2<const DAD*>(s_1, s_2),
                         Seq2<const int*>(s_1 ? s_1->dat() : 0,
                                          s_2 ? s_2->dat() : 0),
                         &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

template<class T>
inline void vecScatter(const Section3<T>& b, const Section3<T>& a,
                       const Section1<int>* s_1, const Section1<int>* s_2,
                       const Section1<int>* s_3,
                       const Section3<int>& msk) {
  VecScatterMsk schedule(&b, &a, sizeof(T),
                         Seq3<const DAD*>(s_1, s_2, s_3),
                         Seq3<const int*>(s_1 ? s_1->dat() : 0,
                                          s_2 ? s_2->dat() : 0,
                                          s_3 ? s_3->dat() : 0),
                         &msk, msk.dat()) ;

  schedule.execute(b.dat(), a.dat()) ;
}

// Reshape

template<class T>
inline void reshape(const Section1<T>& dst, const Section1<T>& src) {
  Reshape schedule(&dst, &src, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void reshape(const Section1<T>& dst, const Section2<T>& src) {
  Reshape schedule(&dst, &src, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void reshape(const Section1<T>& dst, const Section3<T>& src) {
  Reshape schedule(&dst, &src, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void reshape(const Section2<T>& dst, const Section1<T>& src) {
  Reshape schedule(&dst, &src, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void reshape(const Section2<T>& dst, const Section2<T>& src) {
  Reshape schedule(&dst, &src, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void reshape(const Section2<T>& dst, const Section3<T>& src) {
  Reshape schedule(&dst, &src, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void reshape(const Section3<T>& dst, const Section1<T>& src) {
  Reshape schedule(&dst, &src, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void reshape(const Section3<T>& dst, const Section2<T>& src) {
  Reshape schedule(&dst, &src, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void reshape(const Section3<T>& dst, const Section3<T>& src) {
  Reshape schedule(&dst, &src, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void reshape(const Section1<T>& dst, const Section1<T>& src,
                    const Section1<T>& pad) {
  int srcSize = 1 ;
  for(int d = 0 ; d < src.rnk() ; d++)
    srcSize *= src.rng(d).size() ;

  Reshape schedule(&dst, &src, sizeof(T)) ;
  ReshapePad padSchedule(&dst, srcSize, &pad, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
  padSchedule.execute(dst.dat(), pad.dat()) ;
}

template<class T>
inline void reshape(const Section1<T>& dst, const Section2<T>& src,
                    const Section1<T>& pad) {
  int srcSize = 1 ;
  for(int d = 0 ; d < src.rnk() ; d++)
    srcSize *= src.rng(d).size() ;

  Reshape schedule(&dst, &src, sizeof(T)) ;
  ReshapePad padSchedule(&dst, srcSize, &pad, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
  padSchedule.execute(dst.dat(), pad.dat()) ;
}

template<class T>
inline void reshape(const Section1<T>& dst, const Section3<T>& src,
                    const Section1<T>& pad) {
  int srcSize = 1 ;
  for(int d = 0 ; d < src.rnk() ; d++)
    srcSize *= src.rng(d).size() ;

  Reshape schedule(&dst, &src, sizeof(T)) ;
  ReshapePad padSchedule(&dst, srcSize, &pad, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
  padSchedule.execute(dst.dat(), pad.dat()) ;
}

template<class T>
inline void reshape(const Section2<T>& dst, const Section1<T>& src,
                    const Section1<T>& pad) {
  int srcSize = 1 ;
  for(int d = 0 ; d < src.rnk() ; d++)
    srcSize *= src.rng(d).size() ;

  Reshape schedule(&dst, &src, sizeof(T)) ;
  ReshapePad padSchedule(&dst, srcSize, &pad, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
  padSchedule.execute(dst.dat(), pad.dat()) ;
}

template<class T>
inline void reshape(const Section2<T>& dst, const Section2<T>& src,
                    const Section1<T>& pad) {
  int srcSize = 1 ;
  for(int d = 0 ; d < src.rnk() ; d++)
    srcSize *= src.rng(d).size() ;

  Reshape schedule(&dst, &src, sizeof(T)) ;
  ReshapePad padSchedule(&dst, srcSize, &pad, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
  padSchedule.execute(dst.dat(), pad.dat()) ;
}

template<class T>
inline void reshape(const Section2<T>& dst, const Section3<T>& src,
                    const Section1<T>& pad) {
  int srcSize = 1 ;
  for(int d = 0 ; d < src.rnk() ; d++)
    srcSize *= src.rng(d).size() ;

  Reshape schedule(&dst, &src, sizeof(T)) ;
  ReshapePad padSchedule(&dst, srcSize, &pad, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
  padSchedule.execute(dst.dat(), pad.dat()) ;
}

template<class T>
inline void reshape(const Section3<T>& dst, const Section1<T>& src,
                    const Section1<T>& pad) {
  int srcSize = 1 ;
  for(int d = 0 ; d < src.rnk() ; d++)
    srcSize *= src.rng(d).size() ;

  Reshape schedule(&dst, &src, sizeof(T)) ;
  ReshapePad padSchedule(&dst, srcSize, &pad, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
  padSchedule.execute(dst.dat(), pad.dat()) ;
}

template<class T>
inline void reshape(const Section3<T>& dst, const Section2<T>& src,
                    const Section1<T>& pad) {
  int srcSize = 1 ;
  for(int d = 0 ; d < src.rnk() ; d++)
    srcSize *= src.rng(d).size() ;

  Reshape schedule(&dst, &src, sizeof(T)) ;
  ReshapePad padSchedule(&dst, srcSize, &pad, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
  padSchedule.execute(dst.dat(), pad.dat()) ;
}

template<class T>
inline void reshape(const Section3<T>& dst, const Section3<T>& src,
                    const Section1<T>& pad) {
  int srcSize = 1 ;
  for(int d = 0 ; d < src.rnk() ; d++)
    srcSize *= src.rng(d).size() ;

  Reshape schedule(&dst, &src, sizeof(T)) ;
  ReshapePad padSchedule(&dst, srcSize, &pad, sizeof(T)) ;

  schedule.execute(dst.dat(), src.dat()) ;
  padSchedule.execute(dst.dat(), pad.dat()) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `maxval', `minval', `sum', ...                                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
inline T maxval(const Section1<T>& src) {
  T res ;
  Maxval<T> schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

template<class T>
inline T maxval(const Section2<T>& src) {
  T res ;
  Maxval<T> schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

template<class T>
inline T maxval(const Section3<T>& src) {
  T res ;
  Maxval<T> schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

template<class T>
inline T maxval(const Section1<T>& src, const Section1<int>& msk) {
  T res ;
  MaxvalMsk<T> schedule(&src, &msk) ;
  schedule.execute(&res, src.dat(), msk.dat()) ;
  return res ;
}

template<class T>
inline T maxval(const Section2<T>& src, const Section2<int>& msk) {
  T res ;
  MaxvalMsk<T> schedule(&src, &msk) ;
  schedule.execute(&res, src.dat(), msk.dat()) ;
  return res ;
}

template<class T>
inline T maxval(const Section3<T>& src, const Section3<int>& msk) {
  T res ;
  MaxvalMsk<T> schedule(&src, &msk) ;
  schedule.execute(&res, src.dat(), msk.dat()) ;
  return res ;
}


template<class T>
inline T minval(const Section1<T>& src) {
  T res ;
  Minval<T> schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

template<class T>
inline T minval(const Section2<T>& src) {
  T res ;
  Minval<T> schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

template<class T>
inline T minval(const Section3<T>& src) {
  T res ;
  Minval<T> schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

template<class T>
inline T minval(const Section1<T>& src, const Section1<int>& msk) {
  T res ;
  MinvalMsk<T> schedule(&src, &msk) ;
  schedule.execute(&res, src.dat(), msk.dat()) ;
  return res ;
}

template<class T>
inline T minval(const Section2<T>& src, const Section2<int>& msk) {
  T res ;
  MinvalMsk<T> schedule(&src, &msk) ;
  schedule.execute(&res, src.dat(), msk.dat()) ;
  return res ;
}

template<class T>
inline T minval(const Section3<T>& src, const Section3<int>& msk) {
  T res ;
  MinvalMsk<T> schedule(&src, &msk) ;
  schedule.execute(&res, src.dat(), msk.dat()) ;
  return res ;
}


template<class T>
inline T sum(const Section1<T>& src) {
  T res ;
  Sum<T> schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

template<class T>
inline T sum(const Section2<T>& src) {
  T res ;
  Sum<T> schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

template<class T>
inline T sum(const Section3<T>& src) {
  T res ;
  Sum<T> schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

template<class T>
inline T sum(const Section1<T>& src, const Section1<int>& msk) {
  T res ;
  SumMsk<T> schedule(&src, &msk) ;
  schedule.execute(&res, src.dat(), msk.dat()) ;
  return res ;
}

template<class T>
inline T sum(const Section2<T>& src, const Section2<int>& msk) {
  T res ;
  SumMsk<T> schedule(&src, &msk) ;
  schedule.execute(&res, src.dat(), msk.dat()) ;
  return res ;
}

template<class T>
inline T sum(const Section3<T>& src, const Section3<int>& msk) {
  T res ;
  SumMsk<T> schedule(&src, &msk) ;
  schedule.execute(&res, src.dat(), msk.dat()) ;
  return res ;
}

template<class T>
inline T product(const Section1<T>& src) {
  T res ;
  Product<T> schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

template<class T>
inline T product(const Section2<T>& src) {
  T res ;
  Product<T> schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

template<class T>
inline T product(const Section3<T>& src) {
  T res ;
  Product<T> schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

template<class T>
inline T product(const Section1<T>& src, const Section1<int>& msk) {
  T res ;
  ProductMsk<T> schedule(&src, &msk) ;
  schedule.execute(&res, src.dat(), msk.dat()) ;
  return res ;
}

template<class T>
inline T product(const Section2<T>& src, const Section2<int>& msk) {
  T res ;
  ProductMsk<T> schedule(&src, &msk) ;
  schedule.execute(&res, src.dat(), msk.dat()) ;
  return res ;
}

template<class T>
inline T product(const Section3<T>& src, const Section3<int>& msk) {
  T res ;
  ProductMsk<T> schedule(&src, &msk) ;
  schedule.execute(&res, src.dat(), msk.dat()) ;
  return res ;
}

inline int all(const Section1<int>& src) {
  int res ;
  All schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

inline int all(const Section2<int>& src) {
  int res ;
  All schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

inline int all(const Section3<int>& src) {
  int res ;
  All schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

inline int any(const Section1<int>& src) {
  int res ;
  Any schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

inline int any(const Section2<int>& src) {
  int res ;
  Any schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

inline int any(const Section3<int>& src) {
  int res ;
  Any schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

inline int count(const Section1<int>& src) {
  int res ;
  Count schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

inline int count(const Section2<int>& src) {
  int res ;
  Count schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}

inline int count(const Section3<int>& src) {
  int res ;
  Count schedule(&src) ;
  schedule.execute(&res, src.dat()) ;
  return res ;
}


// Dot product variants

inline int dotProduct(const Section1<int>& a, const Section1<int>& b) {
  int res ;
  DotProduct<int, int, int> schedule(&a, &b) ;
  schedule.execute(&res, a.dat(), b.dat()) ;
  return res ;
}

inline float dotProduct(const Section1<int>& a, const Section1<float>& b) {
  float res ;
  DotProduct<float, int, float> schedule(&a, &b) ;
  schedule.execute(&res, a.dat(), b.dat()) ;
  return res ;
}

inline double dotProduct(const Section1<int>& a, const Section1<double>& b) {
  double res ;
  DotProduct<double, int, double> schedule(&a, &b) ;
  schedule.execute(&res, a.dat(), b.dat()) ;
  return res ;
}

inline COMPLEX dotProduct(const Section1<int>& a, const Section1<COMPLEX>& b) {
  COMPLEX res ;
  DotProduct<COMPLEX, int, COMPLEX> schedule(&a, &b) ;
  schedule.execute(&res, a.dat(), b.dat()) ;
  return res ;
}

inline float dotProduct(const Section1<float>& a, const Section1<int>& b) {
  float res ;
  DotProduct<float, float, int> schedule(&a, &b) ;
  schedule.execute(&res, a.dat(), b.dat()) ;
  return res ;
}

inline float dotProduct(const Section1<float>& a, const Section1<float>& b) {
  float res ;
  DotProduct<float, float, float> schedule(&a, &b) ;
  schedule.execute(&res, a.dat(), b.dat()) ;
  return res ;
}

inline double dotProduct(const Section1<float>& a, const Section1<double>& b) {
  double res ;
  DotProduct<double, float, double> schedule(&a, &b) ;
  schedule.execute(&res, a.dat(), b.dat()) ;
  return res ;
}

inline COMPLEX dotProduct(const Section1<float>& a, const Section1<COMPLEX>& b) {
  COMPLEX res ;
  DotProduct<COMPLEX, float, COMPLEX> schedule(&a, &b) ;
  schedule.execute(&res, a.dat(), b.dat()) ;
  return res ;
}

inline double dotProduct(const Section1<double>& a, const Section1<int>& b) {
  double res ;
  DotProduct<double, double, int> schedule(&a, &b) ;
  schedule.execute(&res, a.dat(), b.dat()) ;
  return res ;
}

inline double dotProduct(const Section1<double>& a, const Section1<float>& b) {
  double res ;
  DotProduct<double, double, float> schedule(&a, &b) ;
  schedule.execute(&res, a.dat(), b.dat()) ;
  return res ;
}

inline double dotProduct(const Section1<double>& a, const Section1<double>& b) {
  double res ;
  DotProduct<double, double, double> schedule(&a, &b) ;
  schedule.execute(&res, a.dat(), b.dat()) ;
  return res ;
}

inline COMPLEX dotProduct(const Section1<double>& a,
                          const Section1<COMPLEX>& b) {
  COMPLEX res ;
  DotProduct<COMPLEX, double, COMPLEX> schedule(&a, &b) ;
  schedule.execute(&res, a.dat(), b.dat()) ;
  return res ;
}

inline COMPLEX dotProduct(const Section1<COMPLEX>& a, const Section1<int>& b) {
  COMPLEX res ;
  DotProduct<COMPLEX, COMPLEX, int> schedule(&a, &b) ;
  schedule.execute(&res, a.dat(), b.dat()) ;
  return res ;
}

inline COMPLEX dotProduct(const Section1<COMPLEX>& a,
                          const Section1<float>& b) {
  COMPLEX res ;
  DotProduct<COMPLEX, COMPLEX, float> schedule(&a, &b) ;
  schedule.execute(&res, a.dat(), b.dat()) ;
  return res ;
}

inline COMPLEX dotProduct(const Section1<COMPLEX>& a,
                          const Section1<double>& b) {
  COMPLEX res ;
  DotProduct<COMPLEX, COMPLEX, double> schedule(&a, &b) ;
  schedule.execute(&res, a.dat(), b.dat()) ;
  return res ;
}

inline COMPLEX dotProduct(const Section1<COMPLEX>& a,
                          const Section1<COMPLEX>& b) {
  COMPLEX res ;
  DotProduct<COMPLEX, COMPLEX, COMPLEX> schedule(&a, &b) ;
  schedule.execute(&res, a.dat(), b.dat()) ;
  return res ;
}

inline int boolDotProduct(const Section1<int>& a, const Section1<int>& b) {
  int res ;
  BoolDotProduct schedule(&a, &b) ;
  schedule.execute(&res, a.dat(), b.dat()) ;
  return res ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `maxval', `minval', `sum', dimensional reductions                   //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
inline void sum(const Section1<T>& res,
                const Section2<T>& src, const int dim) {
  SumDim<T> schedule(&res, &src, dim) ;
  schedule.execute(res.dat(), src.dat()) ;
}

template<class T>
inline void sum(const Section2<T>& res,
                const Section3<T>& src, const int dim) {
  SumDim<T> schedule(&res, &src, dim) ;
  schedule.execute(res.dat(), src.dat()) ;
}

template<class T>
inline void sum(const Section1<T>& res,
                const Section2<T>& src, const int dim,
                const Section2<int>& msk) {
  SumDimMsk<T> schedule(&res, &src, dim, &msk) ;
  schedule.execute(res.dat(), src.dat(), msk.dat()) ;
}

template<class T>
inline void sum(const Section2<T>& res,
                const Section3<T>& src, const int dim,
                const Section3<int>& msk) {
  SumDimMsk<T> schedule(&res, &src, dim, &msk) ;
  schedule.execute(res.dat(), src.dat(), msk.dat()) ;
}

template<class T>
inline void product(const Section1<T>& res,
                    const Section2<T>& src, const int dim) {
  ProductDim<T> schedule(&res, &src, dim) ;
  schedule.execute(res.dat(), src.dat()) ;
}

template<class T>
inline void product(const Section2<T>& res,
                    const Section3<T>& src, const int dim) {
  ProductDim<T> schedule(&res, &src, dim) ;
  schedule.execute(res.dat(), src.dat()) ;
}

template<class T>
inline void product(const Section1<T>& res,
                    const Section2<T>& src, const int dim,
                    const Section2<int>& msk) {
  ProductDimMsk<T> schedule(&res, &src, dim, &msk) ;
  schedule.execute(res.dat(), src.dat(), msk.dat()) ;
}

template<class T>
inline void product(const Section2<T>& res,
                    const Section3<T>& src, const int dim,
                    const Section3<int>& msk) {
  ProductDimMsk<T> schedule(&res, &src, dim, &msk) ;
  schedule.execute(res.dat(), src.dat(), msk.dat()) ;
}

template<class T>
inline void maxval(const Section1<T>& res,
                   const Section2<T>& src, const int dim) {
  MaxvalDim<T> schedule(&res, &src, dim) ;
  schedule.execute(res.dat(), src.dat()) ;
}

template<class T>
inline void maxval(const Section2<T>& res,
                   const Section3<T>& src, const int dim) {
  MaxvalDim<T> schedule(&res, &src, dim) ;
  schedule.execute(res.dat(), src.dat()) ;
}

template<class T>
inline void maxval(const Section1<T>& res,
                   const Section2<T>& src, const int dim,
                   const Section2<int>& msk) {
  MaxvalDimMsk<T> schedule(&res, &src, dim, &msk) ;
  schedule.execute(res.dat(), src.dat(), msk.dat()) ;
}

template<class T>
inline void maxval(const Section2<T>& res,
                   const Section3<T>& src, const int dim,
                   const Section3<int>& msk) {
  MaxvalDimMsk<T> schedule(&res, &src, dim, &msk) ;
  schedule.execute(res.dat(), src.dat(), msk.dat()) ;
}

template<class T>
inline void minval(const Section1<T>& res,
                   const Section2<T>& src, const int dim) {
  MinvalDim<T> schedule(&res, &src, dim) ;
  schedule.execute(res.dat(), src.dat()) ;
}

template<class T>
inline void minval(const Section2<T>& res,
                   const Section3<T>& src, const int dim) {
  MinvalDim<T> schedule(&res, &src, dim) ;
  schedule.execute(res.dat(), src.dat()) ;
}

template<class T>
inline void minval(const Section1<T>& res,
                   const Section2<T>& src, const int dim,
                   const Section2<int>& msk) {
  MinvalDimMsk<T> schedule(&res, &src, dim, &msk) ;
  schedule.execute(res.dat(), src.dat(), msk.dat()) ;
}

template<class T>
inline void minval(const Section2<T>& res,
                   const Section3<T>& src, const int dim,
                   const Section3<int>& msk) {
  MinvalDimMsk<T> schedule(&res, &src, dim, &msk) ;
  schedule.execute(res.dat(), src.dat(), msk.dat()) ;
}


inline void all(const Section1<int>& res,
                const Section2<int>& src, const int dim) {
  AllDim schedule(&res, &src, dim) ;
  schedule.execute(res.dat(), src.dat()) ;
}

inline void all(const Section2<int>& res,
                const Section3<int>& src, const int dim) {
  AllDim schedule(&res, &src, dim) ;
  schedule.execute(res.dat(), src.dat()) ;
}

inline void any(const Section1<int>& res,
                const Section2<int>& src, const int dim) {
  AnyDim schedule(&res, &src, dim) ;
  schedule.execute(res.dat(), src.dat()) ;
}

inline void any(const Section2<int>& res,
                const Section3<int>& src, const int dim) {
  AnyDim schedule(&res, &src, dim) ;
  schedule.execute(res.dat(), src.dat()) ;
}

inline void count(const Section1<int>& res,
                  const Section2<int>& src, const int dim) {
  CountDim schedule(&res, &src, dim) ;
  schedule.execute(res.dat(), src.dat()) ;
}

inline void count(const Section2<int>& res,
                  const Section3<int>& src, const int dim) {
  CountDim schedule(&res, &src, dim) ;
  schedule.execute(res.dat(), src.dat()) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `maxloc', `minloc'                                                  //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
inline void maxloc(T* val, int* loc, const Section1<T>& src) {
  Maxloc<T> schedule(&src) ;
  schedule.execute(val, loc, src.dat()) ;
}

template<class T>
inline void maxloc(T* val, int* loc, const Section2<T>& src) {
  Maxloc<T> schedule(&src) ;
  schedule.execute(val, loc, src.dat()) ;
}

template<class T>
inline void maxloc(T* val, int* loc, const Section3<T>& src) {
  Maxloc<T> schedule(&src) ;
  schedule.execute(val, loc, src.dat()) ;
}

template<class T>
inline void maxloc(T* val, int* loc,
                   const Section1<T>& src, const Section1<int>& msk) {
  MaxlocMsk<T> schedule(&src, &msk) ;
  schedule.execute(val, loc, src.dat(), msk.dat()) ;
}

template<class T>
inline void maxloc(T* val, int* loc,
                   const Section2<T>& src, const Section2<int>& msk) {
  MaxlocMsk<T> schedule(&src, &msk) ;
  schedule.execute(val, loc, src.dat(), msk.dat()) ;
}

template<class T>
inline void maxloc(T* val, int* loc,
                   const Section3<T>& src, const Section3<int>& msk) {
  MaxlocMsk<T> schedule(&src, &msk) ;
  schedule.execute(val, loc, src.dat(), msk.dat()) ;
}

template<class T>
inline void minloc(T* val, int* loc, const Section1<T>& src) {
  Minloc<T> schedule(&src) ;
  schedule.execute(val, loc, src.dat()) ;
}

template<class T>
inline void minloc(T* val, int* loc, const Section2<T>& src) {
  Minloc<T> schedule(&src) ;
  schedule.execute(val, loc, src.dat()) ;
}

template<class T>
inline void minloc(T* val, int* loc, const Section3<T>& src) {
  Minloc<T> schedule(&src) ;
  schedule.execute(val, loc, src.dat()) ;
}

template<class T>
inline void minloc(T* val, int* loc,
                   const Section1<T>& src, const Section1<int>& msk) {
  MinlocMsk<T> schedule(&src, &msk) ;
  schedule.execute(val, loc, src.dat(), msk.dat()) ;
}

template<class T>
inline void minloc(T* val, int* loc,
                   const Section2<T>& src, const Section2<int>& msk) {
  MinlocMsk<T> schedule(&src, &msk) ;
  schedule.execute(val, loc, src.dat(), msk.dat()) ;
}

template<class T>
inline void minloc(T* val, int* loc,
                   const Section3<T>& src, const Section3<int>& msk) {
  MinlocMsk<T> schedule(&src, &msk) ;
  schedule.execute(val, loc, src.dat(), msk.dat()) ;
}

template<class T>
inline void maxloc(const Section1<T>& val, const Section1<int>& loc,
                   const Section2<T>& src, const int dim) {
  MaxlocDim<T> schedule(&val, &loc, &src, dim) ;
  schedule.execute(val.dat(), loc.dat(), src.dat()) ;
}

template<class T>
inline void maxloc(const Section2<T>& val, const Section2<int>& loc,
                   const Section3<T>& src, const int dim) {
  MaxlocDim<T> schedule(&val, &loc, &src, dim) ;
  schedule.execute(val.dat(), loc.dat(), src.dat()) ;
}

template<class T>
inline void maxloc(const Section1<T>& val, const Section1<int>& loc,
                   const Section2<T>& src, const int dim,
                   const Section2<int>& msk) {
  MaxlocDimMsk<T> schedule(&val, &loc, &src, dim, &msk) ;
  schedule.execute(val.dat(), loc.dat(), src.dat(), msk.dat()) ;
}

template<class T>
inline void maxloc(const Section2<T>& val, const Section2<int>& loc,
                   const Section3<T>& src, const int dim,
                   const Section3<int>& msk) {
  MaxlocDimMsk<T> schedule(&val, &loc, &src, dim, &msk) ;
  schedule.execute(val.dat(), loc.dat(), src.dat(), msk.dat()) ;
}

template<class T>
inline void minloc(const Section1<T>& val, const Section1<int>& loc,
                   const Section2<T>& src, const int dim) {
  MinlocDim<T> schedule(&val, &loc, &src, dim) ;
  schedule.execute(val.dat(), loc.dat(), src.dat()) ;
}

template<class T>
inline void minloc(const Section2<T>& val, const Section2<int>& loc,
                   const Section3<T>& src, const int dim) {
  MinlocDim<T> schedule(&val, &loc, &src, dim) ;
  schedule.execute(val.dat(), loc.dat(), src.dat()) ;
}

template<class T>
inline void minloc(const Section1<T>& val, const Section1<int>& loc,
                   const Section2<T>& src, const int dim,
                   const Section2<int>& msk) {
  MinlocDimMsk<T> schedule(&val, &loc, &src, dim, &msk) ;
  schedule.execute(val.dat(), loc.dat(), src.dat(), msk.dat()) ;
}

template<class T>
inline void minloc(const Section2<T>& val, const Section2<int>& loc,
                   const Section3<T>& src, const int dim,
                   const Section3<int>& msk) {
  MinlocDimMsk<T> schedule(&val, &loc, &src, dim, &msk) ;
  schedule.execute(val.dat(), loc.dat(), src.dat(), msk.dat()) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `copy'                                                              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
inline void copy(const Section1<T>& dst, const Section1<T>& src) {
  Copy<T> schedule(&dst, &src) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void copy(const Section2<T>& dst, const Section2<T>& src) {
  Copy<T> schedule(&dst, &src) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void copy(const Section3<T>& dst, const Section3<T>& src) {
  Copy<T> schedule(&dst, &src) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `init'                                                              //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
inline void init(const Section1<T>& dst, const T src) {
  Init<T> schedule(&dst) ;

  schedule.execute(dst.dat(), src) ;
}

template<class T>
inline void init(const Section2<T>& dst, const T src) {
  Init<T> schedule(&dst) ;

  schedule.execute(dst.dat(), src) ;
}

template<class T>
inline void init(const Section3<T>& dst, const T src) {
  Init<T> schedule(&dst) ;

  schedule.execute(dst.dat(), src) ;
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `spread'                                                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
inline void spread(const Section2<T>& dst, const Section1<T>& src,
                   const int dim) {
  Spread<T> schedule(&dst, &src, dim) ;

  schedule.execute(dst.dat(), src.dat()) ;
}

template<class T>
inline void spread(const Section3<T>& dst, const Section2<T>& src,
                   const int dim) {
  Spread<T> schedule(&dst, &src, dim) ;

  schedule.execute(dst.dat(), src.dat()) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// `matmul'                                                            //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class S, class T, class U>
inline void matmul(const Section2<S>& res,
                   const Section2<T>& src1, const Section2<U>& src2) {
  
  matmul((DAD*) &res, res.dat(),
         (DAD*) &src1, src1.dat(), (DAD*) &src2, src2.dat()) ;
}


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// Transpose.                                                          //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

template<class T>
void inline transpose(const Section2<T>& dst, const Section2<T>& src) {
  Section2<T> exc(src.rng(1), src.rng(0), src.grp(),
                  src.str(1), src.str(0), src.dat()) ;

  remap(dst, exc) ;
}


// -------------------------------------------------------------
// Things to do
//
//   Eliminate various uses of `Seq' in favour of explicit declaration
//   of local vectors?


