
// Operations on collapsed ranges.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CollapsedRangeRep': collapsed template range.                //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* CollapsedRangeRep :: subrng(const int _extent,
                                      const int _base, const int _stride) {
  return new CollapsedSubRangeRep(_extent, _base, _stride, this, extent) ;
}

RangeRep* CollapsedRangeRep :: ker() const {
  return new DimRangeRep() ;
}

RangeRep* CollapsedRangeRep :: shell() const {
  return (RangeRep*) this ;  // cast away `const'
}

void CollapsedRangeRep :: location(Location* loc, const int glb) const {

  loc->tem = glb ;
  loc->sub = glb ;
  loc->crd = 0 ;
}

void CollapsedRangeRep :: block(const int, Block* blk) const {

  blk->tem_stp = 1 ;
  blk->sub_stp = 1 ;
  blk->glb_stp = 1 ;

  blk->tem_bas = 0 ;
  blk->sub_bas = 0 ;
  blk->glb_bas = 0 ;

  blk->count   = extent ;
}

RangeRep* CollapsedRangeRep :: subker() const {
  return new DimRangeRep() ;
}

int CollapsedRangeRep :: volume() const {return extent ; }

void CollapsedRangeRep :: show() {
  AdlibPrintf("CollapsedRangeRep :: show: "
              "size = %d ker = %p lev = %d dim = %p\n",
              extent, kernel, level, (DimensionRep*) dimension) ; 
}

MEMORY_POOL_DEFN(CollapsedRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `CollapsedSubRangeRep': subrange of a collapsed range.        //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* CollapsedSubRangeRep :: subrng(const int _extent,
                                         const int _base, const int _stride) {
  return new CollapsedSubRangeRep(_extent,
                                  base + stride * _base, stride * _stride,
                                  parent, parSize) ;
}

RangeRep* CollapsedSubRangeRep :: ker() const {
  return new DimRangeRep() ;
}

RangeRep* CollapsedSubRangeRep :: shell() const {
  return (RangeRep*) this ;  // cast away `const'
}

void CollapsedSubRangeRep :: location(Location* loc, const int glb) const {
  const int tem = base + stride * glb ;

  loc->tem = tem ;
  loc->sub = tem ;
  loc->crd = 0 ;
}

void CollapsedSubRangeRep :: block(const int, Block* blk) const {
  blk->tem_stp = stride ;
  blk->sub_stp = stride ;
  blk->glb_stp = 1 ;

  blk->tem_bas = base ;
  blk->sub_bas = base ;
  blk->glb_bas = 0 ;

  blk->count   = extent ;
}

RangeRep* CollapsedSubRangeRep :: subker() const {
  return new DimRangeRep() ;
}

int CollapsedSubRangeRep :: volume() const {return parSize ; }

int CollapsedSubRangeRep :: glb(const int sub, const int) const {
  return (sub - base) / stride ;
}

void CollapsedSubRangeRep :: show() {
  AdlibPrintf("CollapsedSubRangeRep :: show: "
              "size = %d ker = %p lev = %d dim = %p\n",
              extent, kernel, level, (DimensionRep*) dimension) ; 
}

MEMORY_POOL_DEFN(CollapsedSubRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// classes `ExtendedCollapsedRangeRep',                                //
//         `ExtendedCollapsedSubRangeRep':                             //
//     Collapsed range of array with ghost areas.                      //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* ExtendedCollapsedRangeRep :: subrng(const int _extent,
                                              const int _base,
                                              const int _stride) {
  return new ExtendedCollapsedSubRangeRep(_extent, _base, _stride,
                                          this, parSize, wlo, whi) ;
}

int ExtendedCollapsedRangeRep :: disp(const int sub) const {
  return sub + wlo ;
}

int ExtendedCollapsedRangeRep :: volume() const {
  return wlo + parSize + whi ;
}

RangeRep* ExtendedCollapsedSubRangeRep :: subrng(const int _extent,
                                                 const int _base,
                                                 const int _stride) {
  return new ExtendedCollapsedSubRangeRep(_extent,
                                          base + stride * _base,
                                          stride * _stride,
                                          parent, parSize, wlo, whi) ;
}

int ExtendedCollapsedSubRangeRep :: disp(const int sub) const {
  return sub + wlo ;
}

int ExtendedCollapsedSubRangeRep :: volume() const {
  return wlo + parSize + whi ;
}


MEMORY_POOL_DEFN(ExtendedCollapsedRangeRep)

MEMORY_POOL_DEFN(ExtendedCollapsedSubRangeRep)


// -------------------------------------------------------------
// Things to do
//

