
// Operations on level-0 ranges.

#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include "globals.h"
#include "Procs.h"
#include "Range.h"


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `DimRangeRep': process dimension range.                       //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* DimRangeRep :: subrng(const int _extent,
                                const int _base, const int _stride) {
  return new DimSubRangeRep(_extent, _base, _stride, this, dimension) ;
}

RangeRep* DimRangeRep :: ker() const {
  AdlibError("DimRangeRep :: ker: member undefined on this range") ;
  return 0 ;
}

void DimRangeRep :: location(Location* loc, const int glb) const {
  loc->tem = glb ;
  loc->crd = glb ;
  loc->dim = dimension ;
}

int DimRangeRep :: local(int* glb) const {
  *glb = dimension.crd() ;
  return 1 ;
}

RangeRep* DimRangeRep :: subker() const {
  AdlibError("DimRangeRep :: subker: member undefined on this range") ;
  return 0 ;
}

int DimRangeRep :: volume() const {
  return 1 ;
}

int DimRangeRep :: offset(const Location*) const {
  return 0 ;
}

int DimRangeRep :: disp(const int) const {
  AdlibError("DimRangeRep :: disp: member undefined on this range") ;
  return 0 ;
}

int DimRangeRep :: step(const int) const {
  return 0 ;
}

void DimRangeRep :: show() {
  AdlibPrintf("DimRangeRep :: show: size = %d dim = %p\n",
              extent, (DimensionRep*) dimension) ; 
}

MEMORY_POOL_DEFN(DimRangeRep)


// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //
//                                                                     //
// class `DimSubRangeRep': process dimension subrange.                 //
//                                                                     //
// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - //

RangeRep* DimSubRangeRep :: subrng(const int _extent,
                                   const int _base, const int _stride) {
  return new DimSubRangeRep(_extent, base + stride * _base, stride * _stride,
                            parent, dimension) ;
}

RangeRep* DimSubRangeRep :: ker() const {
  AdlibError("DimSubRangeRep :: ker: member undefined on this range") ;
  return 0 ;
}

void DimSubRangeRep :: location(Location* loc, const int glb) const {
  const int crd = base + stride * glb ; 

  loc->tem = crd ;
  loc->crd = crd ;
  loc->dim = dimension ;
}

int DimSubRangeRep :: local(int* glb) const {
  const int offset = dimension.crd() - base ;
  if(offset % stride == 0) {
    *glb = offset / stride ;
    return *glb >= 0 && *glb < extent ;
  }
  else
    return 0 ;
}

RangeRep* DimSubRangeRep :: subker() const {
  AdlibError("DimSubRangeRep :: subker: member undefined on this range") ;
  return 0 ;
}

int DimSubRangeRep :: volume() const {
  return 1 ;
}

int DimSubRangeRep :: offset(const Location*) const {
  return 0 ;
}

int DimSubRangeRep :: disp(const int) const {
  AdlibError("DimSubRangeRep :: disp: member undefined on this range") ;
  return 0 ;
}

int DimSubRangeRep :: step(const int) const {
  return 0 ;
}

int DimSubRangeRep :: glb(const int sub, const int) const {
  return (sub - base) / stride ;
}

void DimSubRangeRep :: show() {
  AdlibPrintf("DimSubRangeRep :: show: size = %d ker = %p lev = %d dim = %p\n",
              extent, kernel, level, (DimensionRep*) dimension) ; 
}

MEMORY_POOL_DEFN(DimSubRangeRep)


// -------------------------------------------------------------
// Things to do
//
//     Currently support ``collapsed dimensions'' through `dim == 0',
//         but don't support subranges of collapsed dimensions.
//         Not a problem?

