
// Subscripting linearized, and ad++ OVERALL macros
// replaced with `LocBlocksIndex'.

#include "ad++.h"

const int N = 128 ;

int main(int argc, char** argv) {
  AdlibInit(argc, argv) ;

  Procs1 p(4) ;
  ON(p) {
    CyclicRange x(N, p.dim(0)) ;
    Range z(N) ;

    // Initialization (unoptimized)

    Array2<float> a(z, x) ;
    Array1<float> b(z) ;  // buffer

    Index m(x) ;
    OVERALL(m) {
      for(int i = 0 ; i < N ; i++)
        if (m >= i)
          a(z(i), m) = i + 1 ;
        else
          a(z(i), m) = m + 1 ;
    } ALLOVER(i) ;

    gprintf("initial a:\n") ;
    aprintf("%7.3f%*N", a, N) ;

    // Do decomposition

    float* a_dat = a.dat() ;
    Stride a0 = a.str(0), a1 = a.str(1) ;

    float* b_dat = b.dat() ;
    Stride b0 = b.str(0) ;

    const int a0_bas = a0.disp(z.bas()) ;
    const int a0_stp = a0.disp_step(z.str()) ;

    const int b0_bas = b0.disp(z.bas()) ;
    const int b0_stp = b0.disp_step(z.str()) ;

    for(int k = 0 ; k < N - 1 ; k++) {

      Location l(x, k) ;
      apgStack.push(&apg) ;
      apg.restrict(l.dim) ;
      if(l.dim.crd() == l.crd) {
        const int a1_off = a1.offset(l) ;

        const float d = sqrt(a_dat [a0_bas + a0_stp * k + a1_off]) ;

        a_dat [a0_bas + a0_stp * k + a1_off] = d ;
        for(int s = k + 1 ; s < N ; s++)
          a_dat [a0_bas + a0_stp * s + a1_off] /= d ;
      }
      apgStack.pop(&apg) ;

      remap(b.sect(z.triplet(k + 1, N - 1)),
            a.sect(z.triplet(k + 1, N - 1), x(k))) ;

      LocBlocksIndex m(x.triplet(k + 1, N - 1)) ;
      apgStack.push(&apg) ;
      apg.restrict(m.dim) ;
      for(m.beginLocBlk() ; m.test() ; m.nextLocBlk()) {
        const int a1_bas = a1.offset(m) ;
        const int a1_stp = a1.step(m) ;

        for(int ml = 0 ; ml < m.count ; ml++) {
	  // NB: `x.idx(m)' is currently expressed as k + 1 + (int) m

	  for(int i = k + 1 + m.glb_bas + m.glb_stp * ml ; i < N ; i++)
	    a_dat [a0_bas + a0_stp * i + a1_bas + a1_stp * ml] -=
		b_dat [b0_bas + b0_stp * i] *
		b_dat [b0_bas + b0_stp * (k + 1 + m.glb_bas + m.glb_stp * ml)] ;
        }
      }
      apgStack.pop(&apg) ;

      gprintf("after iteration %d:\n", k) ;
      aprintf("%7.3f%*N", a, N) ;
    }

    Location l(x, N - 1) ;
    apgStack.push(&apg) ;
    apg.restrict(l.dim) ;
    if(l.dim.crd() == l.crd) {
      const int a1_off = a1.offset(l) ;

      a_dat [a0_bas + a0_stp * (N - 1) + a1_off] = 
          sqrt(a_dat [a0_bas + a0_stp * (N - 1) + a1_off]) ;
    }
    apgStack.pop(&apg) ;

    gprintf("final a:\n") ;
    aprintf("%7.3f%*N", a, N) ;

  } NO(p) ;

  AdlibFinalize() ;
}

